package at.ac.uibk.gitsearch.repository.search;

import at.ac.uibk.gitsearch.repository.jpa.StatisticsRepository;
import at.ac.uibk.gitsearch.service.dto.StatisticsDTO;
import co.elastic.clients.elasticsearch.ElasticsearchClient;
import co.elastic.clients.elasticsearch._types.ElasticsearchException;
import co.elastic.clients.elasticsearch._types.mapping.TypeMapping;
import co.elastic.clients.elasticsearch._types.query_dsl.FieldAndFormat;
import co.elastic.clients.elasticsearch._types.query_dsl.MatchPhrasePrefixQuery;
import co.elastic.clients.elasticsearch._types.query_dsl.MatchQuery;
import co.elastic.clients.elasticsearch._types.query_dsl.Query;
import co.elastic.clients.elasticsearch.cat.AliasesResponse;
import co.elastic.clients.elasticsearch.cat.IndicesResponse;
import co.elastic.clients.elasticsearch.cat.aliases.AliasesRecord;
import co.elastic.clients.elasticsearch.cat.indices.IndicesRecord;
import co.elastic.clients.elasticsearch.core.CountResponse;
import co.elastic.clients.elasticsearch.core.DeleteByQueryResponse;
import co.elastic.clients.elasticsearch.core.DeleteResponse;
import co.elastic.clients.elasticsearch.core.GetResponse;
import co.elastic.clients.elasticsearch.core.SearchResponse;
import co.elastic.clients.elasticsearch.core.search.Hit;
import co.elastic.clients.elasticsearch.indices.DeleteAliasRequest;
import co.elastic.clients.elasticsearch.indices.DeleteAliasResponse;
import co.elastic.clients.elasticsearch.indices.DeleteIndexResponse;
import co.elastic.clients.elasticsearch.indices.update_aliases.AddAction;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.SearchStatisticsDTO;
import org.springframework.stereotype.Repository;

@Repository
@SuppressWarnings("PMD.TooManyMethods")
public class ElasticSearchRepository {

    private final ElasticsearchClient elasticsearchAPIClient;

    private final StatisticsRepository statisticsRepository;

    private static final Logger LOGGER = LogManager.getLogger(ElasticSearchRepository.class);

    protected ElasticSearchRepository(ElasticsearchClient elasticsearchAPIClient, StatisticsRepository statisticsRepository) {
        super();
        this.elasticsearchAPIClient = elasticsearchAPIClient;
        this.statisticsRepository = statisticsRepository;
    }

    /**
     * indexes a SearchResult
     *
     * @param document
     * @throws IOException
     */
    public void index(final SearchResultDTO document) throws IOException {
        index(document, SearchRepositoryConstants.INDEX_METADATA);
    }

    /**
     * indexes a SearchResult
     *
     * @param document
     * @param index    the index
     * @returns the id of the project
     * @throws IOException
     */
    public String index(final SearchResultDTO document, String index) throws IOException {
        this.elasticsearchAPIClient.index(i -> i.index(index).id(document.getExerciseId()).document(document));
        return document.getExerciseId();
    }

    public void waitForIndexingFinished() throws IOException {
        this.elasticsearchAPIClient.indices().refresh(i -> i.index(SearchRepositoryConstants.INDEX_METADATA));
    }

    public List<String> getIdsByProjectId(Long projectId) throws IOException {
        Query.Builder queryBuilder = new Query.Builder();
        queryBuilder.bool(q ->
            q.must(qb -> qb.match(new MatchQuery.Builder().query(projectId).field(SearchRepositoryConstants.METADATA_PROJECT_ID).build()))
        );

        List<FieldAndFormat> idField = Collections.singletonList(FieldAndFormat.of(f -> f.field("_id")));
        final SearchResponse<Object> idResp =
            this.elasticsearchAPIClient.search(
                    d -> d.index(SearchRepositoryConstants.INDEX_METADATA).fields(idField).query(queryBuilder.build()),
                    Object.class
                );

        return idResp.hits().hits().stream().map(Hit::id).collect(Collectors.toList());
    }

    public List<String> getIdsByNamespacePrefix(String namespacePrefix) throws IOException {
        Query.Builder queryBuilder = new Query.Builder();
        queryBuilder.bool(q ->
            q.must(qb ->
                qb.matchPhrasePrefix(
                    new MatchPhrasePrefixQuery.Builder()
                        .field(SearchRepositoryConstants.METADATA_PROJECT_NAMESPACE)
                        .query(namespacePrefix)
                        .build()
                )
            )
        );

        List<FieldAndFormat> idField = Collections.singletonList(FieldAndFormat.of(f -> f.field("_id")));
        final SearchResponse<Object> idResp =
            this.elasticsearchAPIClient.search(
                    d -> d.index(SearchRepositoryConstants.INDEX_METADATA).fields(idField).query(queryBuilder.build()),
                    Object.class
                );

        return idResp.hits().hits().stream().map(Hit::id).collect(Collectors.toList());
    }

    public Long deleteByProjectId(Long projectId) throws IOException {
        Query.Builder queryBuilder = new Query.Builder();
        queryBuilder.bool(q ->
            q.must(qb -> qb.match(new MatchQuery.Builder().query(projectId).field(SearchRepositoryConstants.METADATA_PROJECT_ID).build()))
        );
        final DeleteByQueryResponse deleteByQuery =
            this.elasticsearchAPIClient.deleteByQuery(d -> d.index(SearchRepositoryConstants.INDEX_METADATA).query(queryBuilder.build()));
        return deleteByQuery.deleted();
    }

    public boolean deleteById(String docId) throws IOException {
        final DeleteResponse deleteResp =
            this.elasticsearchAPIClient.delete(d -> d.index(SearchRepositoryConstants.INDEX_METADATA).id(docId));
        return deleteResp.result().name().equals(docId);
    }

    public long countByProjectId(Long projectId) throws IOException {
        Query.Builder queryBuilder = new Query.Builder();
        queryBuilder.bool(q ->
            q.must(qb -> qb.match(new MatchQuery.Builder().query(projectId).field(SearchRepositoryConstants.METADATA_PROJECT_ID).build()))
        );
        final CountResponse count =
            this.elasticsearchAPIClient.count(d -> d.index(SearchRepositoryConstants.INDEX_METADATA).query(queryBuilder.build()));
        return count.count();
    }

    /**
     * returns a document from index metadata
     *
     * @param document
     * @throws ElasticsearchException
     * @throws IOException
     */
    public SearchResultDTO get(final String docId) throws IOException {
        final GetResponse<SearchResultDTO> searchResultDTOResponse =
            this.elasticsearchAPIClient.get(r -> r.index(SearchRepositoryConstants.INDEX_METADATA).id(docId), SearchResultDTO.class);
        return searchResultDTOResponse.source();
    }

    /**
     * return all indices
     *
     * @param document
     * @throws IOException
     */
    public List<IndicesRecord> getIndices() throws IOException {
        final IndicesResponse indicesResponse = this.elasticsearchAPIClient.cat().indices();
        return indicesResponse.valueBody();
    }

    /**
     * return all aliases
     *
     * @param document
     * @throws IOException
     */
    public List<AliasesRecord> getAliases() throws IOException {
        final AliasesResponse indicesResponse = this.elasticsearchAPIClient.cat().aliases();
        return indicesResponse.valueBody();
    }

    /**
     * create an Index
     *
     * @param indexName the name of the index
     * @param mapping   the type mapping
     * @throws IOException
     */
    public void createIndex(final String indexName, final TypeMapping mapping) throws IOException {
        this.elasticsearchAPIClient.indices().create(createIndexRequest -> createIndexRequest.index(indexName).mappings(mapping));
    }

    /**
     * adds/updates searchstatistics to the current doc, if it is existing
     *
     * @param doc
     */
    public void updateSearchStatistics(SearchResultDTO doc) {
        final Optional<StatisticsDTO> statistics = statisticsRepository
            .findByExerciseID(doc.getExerciseId())
            .map(StatisticsDTO::fromEntity);
        statistics.ifPresent(s -> updateSearchStatistics(doc, s));
    }

    public void updateSearchStatistics(SearchResultDTO doc, StatisticsDTO statistics) {
        var ss = new SearchStatisticsDTO();
        ss.setBadgeRewarded(statistics.getBadgeRewarded());
        ss.setDownloads(statistics.getDownloads());
        ss.setViews(statistics.getViews());
        doc.setSearchStatistics(ss);
    }

    /**
     * delete an Index
     *
     * @param indexName the name of the index
     * @throws IOException
     */
    public boolean deleteIndex(final String indexName) throws IOException {
        final DeleteIndexResponse deleteInfo = this.elasticsearchAPIClient.indices().delete(req -> req.index(indexName));
        return deleteInfo.acknowledged();
    }

    /**
     * delete an Alias
     *
     * @param indexName the name of the index
     * @throws IOException
     */
    public boolean deleteAlias(final String aliasName, final String indexName) throws IOException {
        DeleteAliasRequest req = new DeleteAliasRequest.Builder().index(indexName).name(aliasName).build();
        final DeleteAliasResponse deleteInfo = this.elasticsearchAPIClient.indices().deleteAlias(req);
        return deleteInfo.acknowledged();
    }

    /**
     * create an alias for index
     *
     * @param indexName the name of the index
     * @param aliasName the alias name
     * @throws IOException
     */
    public void createAlias(final String aliasName, final String indexName) throws IOException {
        AddAction addAction = new AddAction.Builder().alias(aliasName).index(indexName).build();

        this.elasticsearchAPIClient.indices().updateAliases(req -> req.actions(ab -> ab.add(addAction)));
    }

    /**
     * create an Index
     *
     * @param indexName the name of the index
     * @param mapping   the type mapping
     * @throws IOException
     */
    public void createMetadataIndex(final String indexName) throws IOException {
        createIndex(indexName, getMetaDataConfigDefinition());
    }

    /**
     * returns the meta data mapping
     *
     * @return
     */
    public static InputStream getMetadataIndexMapping() {
        return ElasticSearchRepository.class.getResourceAsStream("./es_metadata.schema.json");
    }

    private TypeMapping getMetaDataConfigDefinition() throws IOException {
        try (InputStream metaDataStream = getMetadataIndexMapping()) {
            var mapper = this.elasticsearchAPIClient._transport().jsonpMapper();
            var parser = mapper.jsonProvider().createParser(metaDataStream);
            return TypeMapping._DESERIALIZER.deserialize(parser, mapper);
        }
    }

    public String createNewMetaDataIndex() throws IOException {
        final List<IndicesRecord> indices = getIndices();

        AtomicInteger next = new AtomicInteger(0);
        while (indices.stream().anyMatch(i -> ("idx_metadata_" + next.intValue()).equals(i.index()))) {
            next.incrementAndGet();
        }
        String indexName = "idx_metadata_" + next;

        createMetadataIndex(indexName);
        return indexName;
    }

    public void createAlias(String indexName) throws IOException {
        final List<AliasesRecord> aliases = getAliases();
        final Optional<AliasesRecord> metaDataAlias = aliases.stream().filter(a -> "metadata".equals(a.alias())).findFirst();

        metaDataAlias.ifPresent(a -> {
            try {
                deleteAlias(a.alias(), a.index());
            } catch (IOException e) {
                LOGGER.error("Cannot delete alias {} for {}", a.alias(), a.index(), e);
            }
        });

        createAlias("metadata", indexName);
    }
}
