package at.ac.uibk.gitsearch.repository.search;

import at.ac.uibk.gitsearch.domain.Authority;
import at.ac.uibk.gitsearch.domain.Statistics;
import at.ac.uibk.gitsearch.domain.util.TreeNode;
import at.ac.uibk.gitsearch.repository.jpa.StatisticsRepository;
import at.ac.uibk.gitsearch.service.dto.AutoCompleteEntry;
import at.ac.uibk.gitsearch.service.dto.StatisticsDTO;
import co.elastic.clients.elasticsearch.ElasticsearchClient;
import co.elastic.clients.elasticsearch._types.query_dsl.BoolQuery;
import co.elastic.clients.elasticsearch._types.query_dsl.ExistsQuery;
import co.elastic.clients.elasticsearch._types.query_dsl.MatchPhrasePrefixQuery;
import co.elastic.clients.elasticsearch._types.query_dsl.MatchQuery;
import co.elastic.clients.elasticsearch._types.query_dsl.MultiMatchQuery;
import co.elastic.clients.elasticsearch._types.query_dsl.PrefixQuery;
import co.elastic.clients.elasticsearch._types.query_dsl.RangeQuery;
import co.elastic.clients.elasticsearch._types.query_dsl.ScriptScoreQuery;
import co.elastic.clients.elasticsearch._types.query_dsl.TermQuery;
import co.elastic.clients.elasticsearch._types.query_dsl.TextQueryType;
import co.elastic.clients.elasticsearch.core.search.Hit;
import co.elastic.clients.json.JsonData;
import co.elastic.clients.transport.rest_client.RestClientTransport;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;
import java.io.IOException;
import java.net.ConnectException;
import java.text.ParseException;
import java.time.Instant;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.ConcurrentModificationException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NoSuchElementException;
import java.util.Optional;
import java.util.Spliterator;
import java.util.Spliterators;
import java.util.StringTokenizer;
import java.util.TreeMap;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Consumer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.DoubleStream;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;
import javax.annotation.PostConstruct;
import javax.ws.rs.NotFoundException;
import liquibase.repackaged.org.apache.commons.text.StringSubstitutor;
import org.apache.commons.lang3.StringUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.codeability.sharing.plugins.api.search.SearchInputDTO;
import org.codeability.sharing.plugins.api.search.SearchOrdering;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.SearchResultsDTO;
import org.codeability.sharing.plugins.api.search.UserProvidedMetadataDTO.Person;
import org.codeability.sharing.plugins.api.search.util.ExerciseId;
import org.elasticsearch.ElasticsearchException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.cache.annotation.Cacheable;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.userdetails.User;
import org.springframework.stereotype.Repository;

@Repository
@SuppressWarnings(
    {
        "PMD.GodClass",
        "PMD.CyclomaticComplexity",
        "PMD.TooManyMethods",
        "PMD.CognitiveComplexity",
        "PMD.NPathComplexity",
        "PMD.ExcessiveClassLength",
        "PMD.AvoidDuplicateLiterals",
    }
)
public class MetaDataRepository {

    private static final String CONNECTION_EXCEPTION_LOGTEMPLATE = "ConnectionException in elastic search: {} ";

    private final ElasticsearchClient elasticsearchAPIClient;

    private static final Logger LOGGER = LogManager.getLogger(MetaDataRepository.class);

    private static final int SCROLL_SIZE = 10;

    public static final String EXERCISE_BY_ID_CACHE = "at.ac.uibk.gitsearch.repository.search.MetaDataRepository.ExerciseByIdCache";

    private static final String SCOPE = "SCOPE";

    @SuppressWarnings("PMD.UnusedPrivateField") // just in case we want to reenable highlighter
    @Value("${application.search.highlight-pre}")
    private String preTags;

    @SuppressWarnings("PMD.UnusedPrivateField") // just in case we want to reenable highlighter
    @Value("${application.search.highlight-post}")
    private String postTags;

    @Autowired
    private StatisticsRepository statisticsRepository;

    /**
     * contains for each Field, a list of Strings mapped to a completion. Together
     * with hit counts E.g. "metadata.creator" -> Kerstin -> {"Kerstin Limbeck" ->
     * 3, "Kerstin Mair" -> 2)}
     */
    private Map<String, Map<String, Map<String, Integer>>> cachedCompletions;

    public MetaDataRepository(ElasticsearchClient elasticsearchAPIClient) {
        this.elasticsearchAPIClient = elasticsearchAPIClient;

        cachedCompletions = new ConcurrentHashMap<>();
        cachedCompletions.put(SearchRepositoryConstants.METADATA_KEYWORDS, new TreeMap<>());
        cachedCompletions.put(SearchRepositoryConstants.METADATA_CREATOR, new TreeMap<>());
        cachedCompletions.put(SearchRepositoryConstants.METADATA_CONTRIBUTOR, new TreeMap<>());
        cachedCompletions.put(SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES, new TreeMap<>());
        cachedCompletions.put(SearchRepositoryConstants.METADATA_FORMATS, new TreeMap<>());
    }

    /**
     * returns all keyword autocompletes for keyWord
     *
     * @param keyWordPrefix
     * @return
     * @throws IOException
     */
    public List<AutoCompleteEntry> getKeywordsAutoComplete(String keyWordPrefix, int max) {
        return getFieldAutoCompletion(keyWordPrefix, SearchRepositoryConstants.METADATA_KEYWORDS, max);
    }

    /**
     * returns all keyword autocompletes for keyWord
     *
     * @param formatPrefix
     * @return
     * @throws IOException
     */
    public List<AutoCompleteEntry> getFormatsAutoComplete(String formatPrefix, int max) {
        return getFieldAutoCompletion(formatPrefix, SearchRepositoryConstants.METADATA_FORMATS, max);
    }

    /**
     * returns all programmingLanguage autocompletes for keyWord
     *
     * @param programmingLanguagePrefix
     * @return
     * @throws IOException
     */
    public List<AutoCompleteEntry> getProgrammingLanguageAutoComplete(String programmingLanguagePrefix, int max) {
        return getFieldAutoCompletion(programmingLanguagePrefix, SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES, max);
    }

    /**
     * returns all creator autocompletes
     *
     * @param creatorPrefix
     * @return
     * @throws IOException
     */
    public List<AutoCompleteEntry> getCreatorAutoComplete(String creatorPrefix, int max) {
        return getFieldAutoCompletion(creatorPrefix, SearchRepositoryConstants.METADATA_CREATOR, max);
    }

    /**
     * returns all contributor autocompletes
     *
     * @param contributorPrefix
     * @return
     * @throws IOException
     */
    public List<AutoCompleteEntry> getContributorAutoComplete(String contributorPrefix, int max) {
        return getFieldAutoCompletion(contributorPrefix, SearchRepositoryConstants.METADATA_CONTRIBUTOR, max);
    }

    /**
     * returns all contributor and creator autocompletes
     *
     * @param contributorPrefix
     * @return
     * @throws IOException
     */
    public List<AutoCompleteEntry> getContributorCreatorAutoComplete(String contributorPrefix, int max) {
        final List<AutoCompleteEntry> contributors = getContributorAutoComplete(contributorPrefix, max);
        final List<AutoCompleteEntry> creatorAutoComplete = getCreatorAutoComplete(contributorPrefix, max);
        // merging
        for (AutoCompleteEntry creator : creatorAutoComplete) {
            contributors
                .stream()
                .filter(entry -> entry.getTarget().equals(creator.getTarget()))
                .findAny()
                .ifPresentOrElse(
                    entry -> {
                        contributors.remove(entry);
                        contributors.add(new AutoCompleteEntry(creator.getTarget(), creator.getHitCount() + entry.getHitCount()));
                    },
                    () -> contributors.add(creator)
                );
        }
        return contributors;
    }

    /**
     * returns a list of hits together with the number of occurences
     *
     * @param prefix        the prefix
     * @param metaDataField the meta data field
     * @param maxCount      the maximum of hits, returned
     * @return
     * @throws IOException
     * @throws JsonProcessingException
     * @throws JsonMappingException
     */
    private List<AutoCompleteEntry> getFieldAutoCompletion(String prefix, final String metaDataField, int maxCount) {
        String lcKeyWordPrefix = prefix.toLowerCase(Locale.getDefault());
        // höllisch kompliziert und höllisch unverständlich mit Streams :-)
        // idea
        // 1. Filtern nach Prefix
        // 2. Zusammenzählen der möglichen Hits (Besser Max statt Sum?)
        // 3. Auf MAX_AUTO_COMPLETION_RESULTS begrenzen und nur die Trefferstrings
        // ausgeben.
        final Stream<Entry<String, Map<String, Integer>>> filteredEntries = cachedCompletions
            .get(metaDataField)
            .entrySet()
            .stream() // Map<String, Map<String, Integer>> -> Stream<Entry<String, Map<String,
            // Integer>>
            .filter(s -> s.getKey().startsWith(lcKeyWordPrefix));
        @SuppressWarnings("null")
        final Map<String, Integer> combinedHits = filteredEntries // s = String -> bool -> Stream<Entry<String,
            // Map<String, Integer>>
            .flatMap(e -> e.getValue().entrySet().stream()) // Stream< Map<String, Integer>> with duplicates
            .collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue, Integer::max));
        return combinedHits
            .entrySet()
            .stream()
            .sorted((e1, e2) -> e2.getValue() - e1.getValue())
            .limit(maxCount)
            .map(s -> new AutoCompleteEntry(s.getKey(), s.getValue()))
            .collect(Collectors.toList()); // Map.Entry<String,
        // Map<String,
        // Integer>>

    }

    @Scheduled(cron = "0 */10 * * * ?") // should run every 10 minutes
    @PostConstruct
    public void updateAutocompletionCache() throws IOException {
        synchronized (this) {
            Map<String, Map<String, Map<String, Integer>>> reloadedCachedCompletions = new ConcurrentHashMap<>();
            reloadedCachedCompletions.put(SearchRepositoryConstants.METADATA_KEYWORDS, new TreeMap<>());
            reloadedCachedCompletions.put(SearchRepositoryConstants.METADATA_CREATOR, new TreeMap<>());
            reloadedCachedCompletions.put(SearchRepositoryConstants.METADATA_CONTRIBUTOR, new TreeMap<>());
            reloadedCachedCompletions.put(SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES, new TreeMap<>());
            reloadedCachedCompletions.put(SearchRepositoryConstants.METADATA_FORMATS, new TreeMap<>());

            try {
                this.getAllResultsAsStream()
                    .forEach(entry -> {
                        if (entry.getMetadata().getKeyword() != null) {
                            for (String keyWord : entry.getMetadata().getKeyword()) {
                                addTo(reloadedCachedCompletions.get(SearchRepositoryConstants.METADATA_KEYWORDS), split(keyWord), keyWord);
                            }
                        }
                        if (entry.getMetadata().getFormat() != null) {
                            for (String format : entry.getMetadata().getFormat()) {
                                addTo(reloadedCachedCompletions.get(SearchRepositoryConstants.METADATA_FORMATS), split(format), format);
                            }
                        }
                        if (entry.getMetadata().getContributor() != null) {
                            for (Person contributor : entry.getMetadata().getContributor()) {
                                addTo(
                                    reloadedCachedCompletions.get(SearchRepositoryConstants.METADATA_CONTRIBUTOR),
                                    split(contributor.getName()),
                                    contributor.getName()
                                );
                            }
                        }
                        if (entry.getMetadata().getCreator() != null) {
                            for (Person creator : entry.getMetadata().getCreator()) {
                                addTo(
                                    reloadedCachedCompletions.get(SearchRepositoryConstants.METADATA_CREATOR),
                                    split(creator.getName()),
                                    creator.getName()
                                );
                            }
                        }
                        if (entry.getMetadata().getProgrammingLanguage() != null) {
                            for (String language : entry.getMetadata().getProgrammingLanguage()) {
                                addTo(
                                    reloadedCachedCompletions.get(SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES),
                                    Collections.singletonList(language),
                                    language
                                );
                            }
                        }
                    });

                reloadedCachedCompletions.values().forEach(this::simplify);
                cachedCompletions = reloadedCachedCompletions;
            } catch (co.elastic.clients.elasticsearch._types.ElasticsearchException | ElasticsearchException | ConnectException ex) {
                // this may happen mainly during testing, when
                // test elasticsearch server is not setup correctly
                LOGGER.error("Cannot update completion Cache", ex);
            } catch (IOException ex) {
                // this may happen mainly during testing, when
                // test elasticsearch server is not setup correctly
                LOGGER.error("IOException: Cannot update completion Cache");
            }
        }
    }

    /**
     * this simplifies completion lists. It combines all found entries with same
     * lowercase representation, and sums them up under the most returned value.
     * (e.g. Java -> 4, java -> 2, JAVA -> 1 is combined to "Java -> 7").
     *
     * @param entries
     */
    public void simplify(Map<String, Map<String, Integer>> entries) {
        entries.forEach((lowerCaseValue, upperCaseValues) -> {
            final Optional<Entry<String, Integer>> max = upperCaseValues.entrySet().stream().max((e1, e2) -> e1.getValue() - e2.getValue());
            max.ifPresent(maxEntry -> {
                int sum = upperCaseValues.entrySet().stream().mapToInt(Entry::getValue).sum();
                entries.put(lowerCaseValue, Collections.singletonMap(maxEntry.getKey(), sum));
            });
        });
    }

    public static String filterDateString(String sourceAsJSON) {
        return sourceAsJSON.replaceAll("[+-]\\d{2}:\\d{2}", "");
    }

    /**
     * returns a preconfigured object mapper
     *
     * @return
     */
    public static ObjectMapper getSearchResultObjectMapper() {
        ObjectMapper objectMapper = new ObjectMapper();
        final JavaTimeModule javaTimeModule = new JavaTimeModule();
        objectMapper.registerModule(javaTimeModule);
        objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
        return objectMapper;
    }

    private static void addTo(Map<String, Map<String, Integer>> completionMap, List<String> tokenList, String target) {
        tokenList.forEach(token -> {
            String lowercaseToken = token.toLowerCase(Locale.getDefault());
            Map<String, Integer> existingTargets = completionMap.get(lowercaseToken);
            if (existingTargets == null) { // none found
                existingTargets = new HashMap<>();
                existingTargets.put(target, 1);
                completionMap.put(lowercaseToken, existingTargets);
            } else {
                final Integer count = existingTargets.get(target);
                if (count == null) {
                    existingTargets.put(target, 1);
                } else {
                    existingTargets.put(target, count + 1);
                }
            }
        });
    }

    private static List<String> split(String s) {
        StringTokenizer st = new StringTokenizer(s, " \t\n\r\f");
        List<String> result = new ArrayList<>();
        while (st.hasMoreTokens()) {
            result.add(st.nextToken());
        }
        return result;
    }

    public List<SearchResultDTO> getAllResources() throws IOException {
        return getAllResultsAsStream().collect(Collectors.toList());
    }

    public SearchResultsDTO searchByEmail(String email, int pageSize, Optional<User> user) throws IOException {
        BoolQuery.Builder queryBuilder = new BoolQuery.Builder();

        BoolQuery.Builder authorBuilder = new BoolQuery.Builder();
        if (email != null) {
            authorBuilder.should(q ->
                q.multiMatch(q2 ->
                    new MultiMatchQuery.Builder()
                        .fields(
                            SearchRepositoryConstants.METADATA_CREATOR_EMAIL,
                            SearchRepositoryConstants.METADATA_CONTRIBUTOR_EMAIL,
                            SearchRepositoryConstants.METADATA_PUBLISHER_EMAIL
                        )
                        .type(TextQueryType.PhrasePrefix)
                        .query(email)
                )
            );
        }

        queryBuilder.must(q -> q.bool(authorBuilder.build()));

        addAuthorizationQueryWithJavaApi(user, queryBuilder);

        try {
            final BoolQuery query = queryBuilder.build();
            // LOGGER.info("ElasticSearch Query using Java Client API:\n{}", query)
            co.elastic.clients.elasticsearch.core.SearchResponse<SearchResultDTO> searchResponse = elasticsearchAPIClient.search(
                search -> search.index(SearchRepositoryConstants.INDEX_METADATA).query(q -> q.bool(query)).from(0).size(pageSize),
                SearchResultDTO.class
            );

            return parseSearchResponseWithJavaAPI(searchResponse);
        } catch (ElasticsearchException ese) {
            LOGGER.warn("Exception in elastic search", ese);
            throw ese;
        } catch (ConnectException ce) {
            @SuppressWarnings("PMD.CloseResource")
            RestClientTransport esTransport = (RestClientTransport) elasticsearchAPIClient._transport();
            final String nodes = esTransport.restClient().getNodes().stream().map(Object::toString).collect(Collectors.joining(", "));
            LOGGER.warn(CONNECTION_EXCEPTION_LOGTEMPLATE, nodes, ce);
            throw ce;
        }
    }

    public static class StreamedSearchResults {

        private final long estimatedSize;
        private final Stream<SearchResultDTO> searchStream;

        protected StreamedSearchResults(long estimatedSize, Stream<SearchResultDTO> searchStream) {
            super();
            this.estimatedSize = estimatedSize;
            this.searchStream = searchStream;
        }

        public long getEstimatedSize() {
            return estimatedSize;
        }

        public Stream<SearchResultDTO> getSearchStream() {
            return searchStream;
        }
    }

    /**
     * Method to stream all exercises.
     *
     * @return the search result page DTO
     * @throws IOException
     */
    public Stream<SearchResultDTO> getAllResultsAsStream() throws IOException {
        final org.codeability.sharing.plugins.api.search.SearchInputMetadataDTO emptySearchMetadata = new org.codeability.sharing.plugins.api.search.SearchInputMetadataDTO(
            null,
            null,
            null,
            null,
            null,
            null
        );
        SearchInputDTO emptySearchQuery = new SearchInputDTO(null, emptySearchMetadata, null, null, null, 0);
        return streamResults(emptySearchQuery, Optional.empty(), true).getSearchStream();
    }

    /**
     * returns the hits from the meta data search index for the input query
     * parameters.
     *
     * @param searchInputDTO          the query parameters.
     * @param user                    the user that queries.
     * @param ignoreUserAuthorization if true, user authorization is ignored
     * @return the search hits.
     * @throws IOException when search index is not available.
     */
    @SuppressWarnings({ "PMD.ExcessiveMethodLength", "PMD.NcssCount" })
    public StreamedSearchResults streamResults(
        final SearchInputDTO searchInputDTO,
        final Optional<User> user,
        boolean ignoreUserAuthorization
    ) throws IOException {
        searchInputDTO.setPageSize(SCROLL_SIZE); // for streaming we set the page size to SCROLL_SIZE
        searchInputDTO.setPage(0); // we force start page to 0!

        final SearchResultsDTO pageDetails = pageDetailsWithJavaApi(searchInputDTO, user, ignoreUserAuthorization);

        final long size = pageDetails.getHitCount();

        Iterator<SearchResultDTO> resultsIterator = new Iterator<>() {
            int count = 0;
            int pagePointer = 0;
            SearchResultsDTO currentPageDetails = pageDetails;

            @Override
            public boolean hasNext() {
                return count < currentPageDetails.getHitCount();
            }

            @Override
            public SearchResultDTO next() {
                if (!hasNext()) {
                    throw new NoSuchElementException();
                }
                if (pagePointer < SCROLL_SIZE) {
                    count++;
                    return currentPageDetails.getSearchResult().get(pagePointer++);
                } else {
                    // load next page
                    searchInputDTO.setPage(searchInputDTO.getPage() + 1);
                    try {
                        currentPageDetails = pageDetailsWithJavaApi(searchInputDTO, user, ignoreUserAuthorization);
                    } catch (IOException e) {
                        throw new ConcurrentModificationException("It seems that scroll result changed during update", e);
                    }
                    pagePointer = 0;
                    return next();
                }
            }
        };

        var resultStream = StreamSupport.stream(Spliterators.spliterator(resultsIterator, (int) size, Spliterator.SIZED), false);
        return new StreamedSearchResults(size, resultStream);
    }

    /**
     * returns the hits from the meta data search index for the input query
     * parameters.
     *
     * @param searchInputDTO the query parameters.
     * @param user           the user that queries.
     * @return the search hits.
     * @throws IOException when search index is not available.
     */
    @SuppressWarnings({ "PMD.ExcessiveMethodLength", "PMD.NcssCount" })
    public SearchResultsDTO pageDetailsWithJavaApi(SearchInputDTO searchInputDTO, Optional<User> user) throws IOException {
        return pageDetailsWithJavaApi(searchInputDTO, user, false);
    }

    /**
     * returns the hits from the meta data search index for the input query
     * parameters.
     *
     * @param searchInputDTO          the query parameters.
     * @param user                    the user that queries.
     * @param ignoreUserAuthorization ignores the authorization of the user
     * @return the search hits.
     * @throws IOException when search index is not available.
     */
    @SuppressWarnings({ "PMD.ExcessiveMethodLength", "PMD.NcssCount" })
    public SearchResultsDTO pageDetailsWithJavaApi(SearchInputDTO searchInputDTO, Optional<User> user, boolean ignoreUserAuthorization)
        throws IOException {
        final int pageSize = searchInputDTO.getPageSize();
        final int from = (searchInputDTO.getPage()) * pageSize;

        try {
            final BoolQuery.Builder boolQueryBuilder = createBoolQueryBuilder(searchInputDTO, user, ignoreUserAuthorization);
            final ScriptScoreQuery.Builder scoreQueryBuilder = createScoreQueryBuilder(searchInputDTO, boolQueryBuilder.build());
            co.elastic.clients.elasticsearch.core.SearchResponse<SearchResultDTO> searchResponse = elasticsearchAPIClient.search(
                search ->
                    search
                        .index(SearchRepositoryConstants.INDEX_METADATA)
                        .query(q -> q.scriptScore(scoreQueryBuilder.build()))
                        .from(from)
                        .size(pageSize),
                SearchResultDTO.class
            );

            return parseSearchResponseWithJavaAPI(searchResponse);
        } catch (ElasticsearchException ese) {
            LOGGER.warn("Exception in elastic search", ese);
            throw ese;
        } catch (ConnectException ce) {
            @SuppressWarnings("PMD.CloseResource")
            RestClientTransport esTransport = (RestClientTransport) elasticsearchAPIClient._transport();
            final String nodes = esTransport.restClient().getNodes().stream().map(Object::toString).collect(Collectors.joining(", "));
            LOGGER.warn(CONNECTION_EXCEPTION_LOGTEMPLATE, nodes, ce);

            throw ce;
        }
    }

    private ScriptScoreQuery.Builder createScoreQueryBuilder(SearchInputDTO searchInputDTO, final BoolQuery query) {
        final SearchOrdering searchOrdering = searchInputDTO.getOrdering();
        String sortingScriptTemplate =
            "double boost = 0;" +
            "if (doc.containsKey('${badgeRewarded}') && doc['${badgeRewarded}'].size() > 0) {" +
            "    boost = boost + (doc['${badgeRewarded}'].value ? 1.0 : 0.0) * params.factorBadge;" +
            "}" +
            "if (doc.containsKey('${downloads}') && doc['${downloads}'].size() > 0) {" +
            "    boost = boost + doc['${downloads}'].value * params.factorDownloads;" +
            "}" +
            "if (doc.containsKey('${views}') && doc['${views}'].size() > 0) {" +
            "    boost = boost + doc['${views}'].value * params.factorViews;" +
            "}" +
            "return boost + _score;";

        Map<String, String> templateValueMapping = Map.of(
            "badgeRewarded",
            SearchRepositoryConstants.SEARCHSTATISTICS_BADGEREWARDED,
            "downloads",
            SearchRepositoryConstants.SEARCHSTATISTICS_DOWNLOADS,
            "views",
            SearchRepositoryConstants.SEARCHSTATISTICS_VIEWS
        );

        double maxFactor =
            DoubleStream
                .of(searchOrdering.getFactorBadge(), searchOrdering.getFactorDownloads(), searchOrdering.getFactorViews())
                .max()
                .getAsDouble() *
            0.2;

        return new ScriptScoreQuery.Builder()
            .script(sb ->
                sb.inline(isb ->
                    isb
                        .lang("painless")
                        .source(StringSubstitutor.replace(sortingScriptTemplate, templateValueMapping))
                        .params("factorBadge", JsonData.of(searchOrdering.getFactorBadge() / maxFactor))
                        .params("factorDownloads", JsonData.of(searchOrdering.getFactorDownloads() / maxFactor))
                        .params("factorViews", JsonData.of(searchOrdering.getFactorViews() / maxFactor))
                )
            )
            .query(q -> q.bool(query));
    }

    /**
     * creates the query (builder) for the searchInput and the user authorization
     *
     * @param searchInputDTO          the search input
     * @param user                    the user (optional)
     * @param ignoreUserAuthorization if true, the user authorization is ignored
     * @return
     */
    @SuppressWarnings("PMD.ExcessiveMethodLength")
    private BoolQuery.Builder createBoolQueryBuilder(SearchInputDTO searchInputDTO, Optional<User> user, boolean ignoreUserAuthorization) {
        BoolQuery.Builder queryBuilder = new BoolQuery.Builder();

        forEachTerm(
            searchInputDTO.getFulltextQuery(),
            term ->
                queryBuilder.must(q ->
                    q.multiMatch(
                        new MultiMatchQuery.Builder()
                            .query(term)
                            .fields(
                                SearchRepositoryConstants.METADATA_DESCRIPTION,
                                SearchRepositoryConstants.METADATA_KEYWORDS,
                                SearchRepositoryConstants.METADATA_TITLE
                            )
                            .type(TextQueryType.PhrasePrefix)
                            .build()
                    )
                )
        );

        BoolQuery.Builder plBuilder = new BoolQuery.Builder();
        forEachTerm(
            searchInputDTO.getMetadata().getProgrammingLanguage(),
            term ->
                plBuilder.should(q ->
                    q.prefix(
                        new PrefixQuery.Builder()
                            .value(term)
                            .caseInsensitive(true)
                            .field(SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES)
                            .build()
                    )
                ),
            () -> queryBuilder.must(q -> q.bool(plBuilder.build()))
        );

        BoolQuery.Builder keywordBuilder = new BoolQuery.Builder();
        forEachTerm(
            searchInputDTO.getMetadata().getKeyword(),
            term ->
                keywordBuilder.should(q ->
                    q.matchPhrasePrefix(
                        new MatchPhrasePrefixQuery.Builder().query(term).field(SearchRepositoryConstants.METADATA_KEYWORDS).build()
                    )
                ),
            () -> queryBuilder.must(q -> q.bool(keywordBuilder.build()))
        );

        BoolQuery.Builder formatBuilder = new BoolQuery.Builder();
        forEachTerm(
            searchInputDTO.getMetadata().getFormat(),
            term ->
                formatBuilder.should(q ->
                    q.prefix(
                        new PrefixQuery.Builder()
                            .value(term)
                            .field(SearchRepositoryConstants.METADATA_FORMATS)
                            .caseInsensitive(true)
                            .build()
                    )
                ),
            () -> queryBuilder.must(q -> q.bool(formatBuilder.build()))
        );

        String parentId = searchInputDTO.getMetadata().getParentId();
        SearchResultDTO parentExercise = null;
        if (!StringUtils.isEmpty(parentId)) {
            try {
                parentExercise = getExerciseById(parentId, user);
            } catch (NotFoundException e) {
                String stripped = convertToOldFormat(parentId);
                if (stripped.equals(parentId)) {
                    try {
                        parentExercise = getExerciseById(stripped, user);
                    } catch (NotFoundException e1) {
                        LOGGER.warn("Cannot find exercise for: {}", parentId);
                    }
                }
            }
            if (parentExercise == null || parentExercise.getFile().getChildren() == null) {
                // just a very lazy workaround
                return null;
            } else {
                String[] childrenIds = parentExercise.getFile().getChildren();
                BoolQuery.Builder collectionBuilder = new BoolQuery.Builder();
                for (String childId : childrenIds) {
                    collectionBuilder.should(q ->
                        q.term((new TermQuery.Builder()).value(childId).field(SearchRepositoryConstants.EXERCISE_ID).build())
                    );
                    String previousId = convertToOldFormat(childId);
                    if (!previousId.equals(childId)) {
                        collectionBuilder.should(q ->
                            q.term((new TermQuery.Builder()).value(previousId).field(SearchRepositoryConstants.EXERCISE_ID).build())
                        );
                    }
                }
                queryBuilder.must(q -> q.bool(collectionBuilder.build()));
            }
        }

        if (searchInputDTO.getMetadata().getAuthor() != null) {
            queryBuilder.must(q1 ->
                q1.bool(q2 ->
                    q2.should(q ->
                        q.multiMatch(
                            new MultiMatchQuery.Builder()
                                .query(searchInputDTO.getMetadata().getAuthor())
                                .type(TextQueryType.PhrasePrefix)
                                .fields(
                                    SearchRepositoryConstants.METADATA_CREATOR,
                                    SearchRepositoryConstants.METADATA_CONTRIBUTOR,
                                    SearchRepositoryConstants.METADATA_PUBLISHER
                                )
                                .build()
                        )
                    )
                )
            );
        }

        if (searchInputDTO.getMetadata().getNaturalLanguage() != null && !searchInputDTO.getMetadata().getNaturalLanguage().isEmpty()) {
            BoolQuery.Builder nlBuilder = new BoolQuery.Builder();

            searchInputDTO
                .getMetadata()
                .getNaturalLanguage()
                .forEach(language ->
                    nlBuilder.should(q ->
                        q.match(new MatchQuery.Builder().query(language).field(SearchRepositoryConstants.METADATA_LANGUAGE).build())
                    )
                );
            queryBuilder.must(q -> q.bool((nlBuilder.build())));
        }

        forEachTerm(
            searchInputDTO.getMetadata().getLicense(),
            term ->
                queryBuilder.must(q ->
                    q.prefix(new PrefixQuery.Builder().value(term).field(SearchRepositoryConstants.METADATA_LICENSE).build())
                )
        );

        if (!ignoreUserAuthorization) {
            addAuthorizationQueryWithJavaApi(user, queryBuilder);
        }

        return queryBuilder;
    }

    /**
     * strips brackets from exercise Id. Just for temporary use. Will not be
     * necessary, when we change to bracketed ids in general.
     *
     * @param exerciseId
     * @return
     */
    public String convertToOldFormat(String exerciseId) {
        Pattern newParser = Pattern.compile("^\\[(\\d+)\\](.*)$");
        Matcher matcher = newParser.matcher(exerciseId.trim());
        if (matcher.matches()) {
            String projectId = matcher.group(1);
            String path = matcher.group(2);
            if (StringUtils.isEmpty(path)) {
                return projectId;
            } else {
                return projectId + ":" + path;
            }
        }
        return exerciseId;
    }

    /**
     * this adds extra authorization requirements. It currently supports directly
     * the group membership in the hierarchy and the new access management by shared
     * groups and explicit members.
     *
     * @param user
     * @param queryBuilder
     */
    private void addAuthorizationQueryWithJavaApi(Optional<User> user, BoolQuery.Builder queryBuilder) {
        // Authorization restrictions
        // the boost value of 0 tells elastic to exclude these restrictions from the scoring process
        final TermQuery.Builder simplePublicQuery = new TermQuery.Builder()
            .value("public")
            .field(SearchRepositoryConstants.PROJECT_VISIBILITY)
            .boost(0.0f);
        final ExistsQuery.Builder publicVisibilityQuery = new ExistsQuery.Builder()
            .field(SearchRepositoryConstants.METADATA_PUBLICVISIBILITY)
            .boost(0.0f);
        final BoolQuery.Builder publicQuery = new BoolQuery.Builder()
            .should(q -> q.term(simplePublicQuery.build()))
            .should(q -> q.exists(publicVisibilityQuery.build()));

        if (user.isEmpty()) {
            queryBuilder.must(q -> q.bool(publicQuery.build()));
        } else {
            final AtomicBoolean hasAuthQueries = new AtomicBoolean(false);
            final Collection<GrantedAuthority> authorities = user.get().getAuthorities();
            final BoolQuery.Builder authQuery = new BoolQuery.Builder().boost(0.0f);
            // old membership management (can be eliminated some day)
            final Stream<PrefixQuery.Builder> prefixAuthQueries = authorities
                .stream()
                .filter(auth -> !Authority.isStandardRole(auth.getAuthority()))
                .filter(auth -> !auth.getAuthority().startsWith(SCOPE))
                .map(auth -> new PrefixQuery.Builder().field(SearchRepositoryConstants.PROJECT_NAMESPACE).value(auth.getAuthority()));
            prefixAuthQueries.forEach(pq -> {
                hasAuthQueries.set(true);
                authQuery.should(q -> q.prefix(pq.build()));
            });

            // new membership management
            final Stream<TermQuery.Builder> prefixAuthQueries2 = authorities
                .stream()
                .filter(auth -> !Authority.isStandardRole(auth.getAuthority()))
                .filter(auth -> !auth.getAuthority().startsWith(SCOPE))
                .map(auth -> new TermQuery.Builder().field(SearchRepositoryConstants.PROJECT_GROUPS).value(auth.getAuthority()));
            prefixAuthQueries2.forEach(paq -> {
                hasAuthQueries.set(true);
                authQuery.should(q -> q.term(paq.build()));
            });

            final String email = user.get().getUsername().toLowerCase(Locale.getDefault());
            if (StringUtils.isNotEmpty(email)) {
                hasAuthQueries.set(true);
                authQuery.should(q -> q.term(new TermQuery.Builder().field(SearchRepositoryConstants.PROJECT_USERS).value(email).build()));
            }

            if (hasAuthQueries.get()) {
                queryBuilder.must(q1 ->
                    q1.bool(new BoolQuery.Builder().should(q -> q.bool(authQuery.build())).should(q -> q.bool(publicQuery.build())).build())
                );
            } else {
                queryBuilder.must(q -> q.bool(publicQuery.build()));
            }
        }
    }

    private SearchResultsDTO parseSearchResponseWithJavaAPI(
        final co.elastic.clients.elasticsearch.core.SearchResponse<SearchResultDTO> searchResponse
    ) {
        final Double maxScoreD = searchResponse.hits().maxScore();
        final double maxScore = maxScoreD == null ? 0.0d : maxScoreD;

        final List<SearchResultDTO> searchResults = new ArrayList<>();
        for (Hit<SearchResultDTO> searchHit : searchResponse.hits().hits()) {
            final Double scoreD = searchHit.score();
            double hitScore = scoreD == null ? 0.0 : scoreD;
            final SearchResultDTO parsedSearchHit = searchHit.source();
            if (parsedSearchHit == null) {
                LOGGER.warn("received null search hit?");
                continue;
            }

            parsedSearchHit.setExerciseId(searchHit.id());
            int ranking5 = 1;
            if (maxScore > 0.0f) {
                ranking5 = (int) ((hitScore * 5.0) / maxScore + 0.5f);
            }
            parsedSearchHit.setRanking5(ranking5);
            searchResults.add(parsedSearchHit);
        }

        co.elastic.clients.elasticsearch.core.search.TotalHits totalHits = searchResponse.hits().total();
        long hitCount = 0;
        if (totalHits != null) {
            hitCount = totalHits.value();
        }
        final SearchResultsDTO results = new SearchResultsDTO();
        results.setHitCount(hitCount);
        results.setSearchResult(searchResults);

        final List<Statistics> statistics = statisticsRepository.findStatisticsByExerciseIds(
            results.getSearchResult().stream().map(sr -> sr.getExerciseId()).collect(Collectors.toList())
        );
        results
            .getSearchResult()
            .forEach(sr ->
                statistics
                    .stream()
                    .filter(stat -> stat.getExerciseID().equals(sr.getExerciseId()))
                    .findAny()
                    .ifPresent(stat -> sr.setSearchStatistics(StatisticsDTO.fromEntityToSearchStatistics(stat)))
            );

        return results;
    }

    /** Helper **/
    private static void forEachTerm(String searchTerm, Consumer<String> exec) {
        forEachTerm(searchTerm, exec, () -> {});
    }

    /** Helper **/
    private static void forEachTerm(String searchTerm, Consumer<String> exec, Runnable whenAtLeastOne) {
        if (searchTerm == null) {
            return;
        }
        StringTokenizer st = new StringTokenizer(searchTerm);

        boolean atLeastOne = false;
        while (st.hasMoreTokens()) {
            atLeastOne = true;
            exec.accept(st.nextToken());
        }
        if (atLeastOne) {
            whenAtLeastOne.run();
        }
    }

    public SearchResultsDTO getExercisesById(Stream<String> exerciseIds, Optional<User> user, int page, int pageSize) {
        return getExercisesById(exerciseIds, user, page, pageSize, null);
    }

    /**
     * returns all exercises, referred by their ids, changed after since (if not
     * null)
     *
     * @param exerciseIds a stream of exercise ids
     * @param user        the user: needed for authorization
     * @param page        the page
     * @param pageSize    the page size
     * @param since       an instant since when the exercise was last changed (or
     *                    null, if not relevant)
     * @return
     */
    @SuppressWarnings("PMD.PreserveStackTrace")
    public SearchResultsDTO getExercisesById(Stream<String> exerciseIds, Optional<User> user, int page, int pageSize, Instant since) {
        BoolQuery.Builder idQueryBuilder = new BoolQuery.Builder();

        AtomicInteger idCounter = new AtomicInteger(0);
        exerciseIds.forEach(exerciseId -> {
            idCounter.incrementAndGet();
            idQueryBuilder.should(q ->
                q.match(q2 -> new MatchQuery.Builder().field(SearchRepositoryConstants.EXERCISE_ID).query(exerciseId))
            );
        });

        // Böse Falle: Wenn exerciseIds empty, dann wird alles zurückgegeben!
        if (idCounter.get() == 0) {
            final SearchResultsDTO results = new SearchResultsDTO();
            results.setHitCount(0L);
            results.setSearchResult(new ArrayList<>());
            return results;
        }
        BoolQuery.Builder queryBuilder = new BoolQuery.Builder();

        queryBuilder.must(q -> q.bool(idQueryBuilder.build()));

        if (since != null) {
            // Noch ein böse Falle! JsonData.of(since) konvertiert nicht in Zeitstring
            // sondern in
            // ein Array of Strings
            String sinceString = DateTimeFormatter.ISO_INSTANT.format(since);
            // "2023-03-15T08:32:52
            queryBuilder.must(q ->
                q.range(
                    new RangeQuery.Builder().field(SearchRepositoryConstants.PROJECT_LASTACTIVITYAT).gte(JsonData.of(sinceString)).build()
                )
            );
        }

        addAuthorizationQueryWithJavaApi(user, queryBuilder);

        try {
            final BoolQuery query = queryBuilder.build();
            LOGGER.info("ElasticSearch Query using Java Client API:\n{}", query);
            co.elastic.clients.elasticsearch.core.SearchResponse<SearchResultDTO> searchResponse = elasticsearchAPIClient.search(
                search ->
                    search.index(SearchRepositoryConstants.INDEX_METADATA).query(q -> q.bool(query)).from(page * pageSize).size(pageSize),
                SearchResultDTO.class
            );

            return parseSearchResponseWithJavaAPI(searchResponse);
        } catch (ElasticsearchException ese) {
            LOGGER.warn("Exception in elastic search", ese);
            throw ese;
        } catch (IOException ce) {
            @SuppressWarnings("PMD.CloseResource")
            RestClientTransport esTransport = (RestClientTransport) elasticsearchAPIClient._transport();
            final String nodes = esTransport.restClient().getNodes().stream().map(Object::toString).collect(Collectors.joining(", "));
            LOGGER.warn(CONNECTION_EXCEPTION_LOGTEMPLATE, nodes, ce);
            throw new ElasticsearchException("IOException", ce);
        }
    }

    /**
     * Searches for an exercise by id just a convenience method
     *
     * @param exerciseId the parsed exercise id
     * @param user       the user (if authenticated)
     * @return search result
     * @throws NotFoundException if not found
     */
    @Cacheable(value = EXERCISE_BY_ID_CACHE, key = "((#user.present)?#user.get().username:'') + #exerciseId.toString()")
    public SearchResultDTO getExerciseById(ExerciseId exerciseId, Optional<User> user) {
        return getExerciseById(exerciseId.toString(), user);
    }

    /**
     * calculates recursively the list of elements in the collection. TODO make
     * more efficient by looking up all children in one go.
     *
     * @param exerciseId
     * @param user
     * @return
     */
    @SuppressWarnings("PMD.CognitiveComplexity")
    public TreeNode<SearchResultDTO> getExerciseCollectionTree(ExerciseId exerciseId, Optional<User> user) {
        SearchResultDTO root = getExerciseById(exerciseId, user);
        if (root == null) {
            LOGGER.debug("Exercise with id {} not found", exerciseId);
            return null;
        }
        LOGGER.debug("Loaded Exercise with id {}: ({})", exerciseId, root);
        TreeNode<SearchResultDTO> result = new TreeNode<>(root);
        final String[] children = root.getFile().getChildren();
        if (children != null) {
            for (String childId : children) {
                try {
                    final ExerciseId parsedChildId = ExerciseId.fromString(childId);
                    // only apply to all children in the same project
                    if (parsedChildId.isDescendantOf(exerciseId)) {
                        TreeNode<SearchResultDTO> childResult = getExerciseCollectionTree(parsedChildId, user);
                        if (childResult != null) {
                            result.addChild(childResult);
                        }
                    }
                } catch (ParseException e) {
                    LOGGER.warn("Cannot parse {}", exerciseId);
                    return null;
                }
            }
        }
        return result;
    }

    /**
     * Searches for an exercise by id
     *
     * @param exerciseId exercise id as String
     * @param user       the user (if authenticated)
     * @return search result
     * @throws NotFoundException if not found
     */
    @Cacheable(value = EXERCISE_BY_ID_CACHE, key = "((#user.present)?#user.get().username:'') + #exerciseId.toString()")
    public SearchResultDTO getExerciseById(String exerciseId, Optional<User> user) {
        String bracketedExerciseId = exerciseId;
        if (!bracketedExerciseId.startsWith("[")) { // just for backward compatibility
            bracketedExerciseId = "[" + exerciseId + "]";
        }
        final String finalBracketedExerciseId = bracketedExerciseId;

        BoolQuery.Builder queryBuilder = new BoolQuery.Builder();

        queryBuilder.must(q ->
            q.match(q2 -> new MatchQuery.Builder().field(SearchRepositoryConstants.EXERCISE_ID).query(finalBracketedExerciseId))
        );

        addAuthorizationQueryWithJavaApi(user, queryBuilder);

        try {
            final BoolQuery query = queryBuilder.build();
            // LOGGER.info("ElasticSearch Query using Java Client API:\n{}", query)
            co.elastic.clients.elasticsearch.core.SearchResponse<SearchResultDTO> searchResponse = elasticsearchAPIClient.search(
                search -> search.index(SearchRepositoryConstants.INDEX_METADATA).query(q -> q.bool(query)).from(0).size(1),
                SearchResultDTO.class
            );

            co.elastic.clients.elasticsearch.core.search.TotalHits totalHits = searchResponse.hits().total();
            long hitCount = 0;
            if (totalHits != null) {
                hitCount = totalHits.value();
            }
            if (hitCount > 1) {
                LOGGER.warn("Found {} hits for exercise id {}, expected only one!", hitCount, exerciseId);
            }

            if (hitCount > 0) {
                final SearchResultsDTO parsedHits = parseSearchResponseWithJavaAPI(searchResponse);
                return parsedHits.getSearchResult().get(0);
            }
        } catch (ElasticsearchException ese) {
            LOGGER.warn("Exception in elastic search", ese);
            throw ese;
        } catch (IOException ce) {
            @SuppressWarnings("PMD.CloseResource")
            final RestClientTransport esTransport = (RestClientTransport) elasticsearchAPIClient._transport();
            final String nodes = esTransport.restClient().getNodes().stream().map(Object::toString).collect(Collectors.joining(", "));
            LOGGER.warn(CONNECTION_EXCEPTION_LOGTEMPLATE, nodes, ce);
            throw new NotFoundException("Exercise with id " + exerciseId + " not found", ce);
        }
        throw new NotFoundException("Exercise with id " + exerciseId + " not found");
    }
}
