package at.ac.uibk.gitsearch.service.gitlab_events;

import at.ac.uibk.gitsearch.repository.gitlab.GitLabRepository;
import at.ac.uibk.gitsearch.repository.search.ElasticSearchRepository;
import at.ac.uibk.gitsearch.repository.search.SearchRepositoryConstants;
import at.ac.uibk.gitsearch.service.MailService;
import at.ac.uibk.gitsearch.service.util.SlashPath;
import at.ac.uibk.gitsearch.service.vocabulary.VocabularyService;
import at.ac.uibk.gitsearch.service.vocabulary.VocabularyService.ValidationResultDTO;
import ch.qos.logback.classic.LoggerContext;
import ch.qos.logback.classic.encoder.PatternLayoutEncoder;
import ch.qos.logback.classic.spi.ILoggingEvent;
import ch.qos.logback.core.OutputStreamAppender;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.JsonParser;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.JsonToken;
import com.fasterxml.jackson.databind.DeserializationContext;
import com.fasterxml.jackson.databind.JsonDeserializer;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.annotation.JsonDeserialize;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;
import com.networknt.schema.JsonSchema;
import com.networknt.schema.JsonSchemaFactory;
import java.io.IOException;
import java.io.InputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.nio.charset.StandardCharsets;
import java.text.ParseException;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Base64;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.Objects;
import java.util.Optional;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import javax.annotation.PostConstruct;
import javax.el.MethodNotFoundException;
import org.apache.commons.lang3.StringUtils;
import org.codeability.sharing.plugins.api.search.GitProjectDTO;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.SearchResultDTO.MetadataFile;
import org.codeability.sharing.plugins.api.search.UserProvidedMetadataDTO;
import org.codeability.sharing.plugins.api.search.UserProvidedMetadataDTO.InteractivityTypeXXX;
import org.codeability.sharing.plugins.api.search.UserProvidedMetadataDTO.PublicVisibility;
import org.gitlab4j.api.GitLabApi;
import org.gitlab4j.api.GitLabApiException;
import org.gitlab4j.api.models.Branch;
import org.gitlab4j.api.models.Commit;
import org.gitlab4j.api.models.Member;
import org.gitlab4j.api.models.Project;
import org.gitlab4j.api.models.ProjectSharedGroup;
import org.gitlab4j.api.models.RepositoryFile;
import org.gitlab4j.api.models.TreeItem;
import org.gitlab4j.api.models.User;
import org.gitlab4j.api.systemhooks.GroupSystemHookEvent;
import org.gitlab4j.api.systemhooks.ProjectSystemHookEvent;
import org.gitlab4j.api.systemhooks.PushSystemHookEvent;
import org.gitlab4j.api.systemhooks.SystemHookEvent;
import org.gitlab4j.api.systemhooks.TeamMemberSystemHookEvent;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.thymeleaf.context.Context;
import org.thymeleaf.spring5.SpringTemplateEngine;
import tech.jhipster.config.JHipsterProperties;

/**
 * Service to handle gitlab events
 * <p>
 */
@Service
@SuppressWarnings(
    {
        "PMD.ExcessivePublicCount",
        "PMD.NonStaticInitializer",
        "PMD.ExcessiveClassLength",
        "PMD.TooManyMethods",
        "PMD.AvoidDuplicateLiterals",
    }
)
public class GitlabEventService {

    @Autowired
    private GitLabRepository gitLabRepository;

    private GitLabApi adminGitLabApi;

    @Autowired
    private VocabularyService vocabularyService;

    @Autowired
    public ElasticSearchRepository esRepository;

    @Autowired
    private MailService mailService;

    @Autowired
    private SpringTemplateEngine templateEngine;

    @Autowired
    private JHipsterProperties jHipsterProperties;

    protected static final Pattern collectionContentPattern = Pattern.compile("^(http[^\\[]+)?(\\[(\\d*)\\])?(.+)?$");

    static final Logger logger = LoggerFactory.getLogger(GitlabEventService.class);

    @SuppressWarnings("PMD.DoNotUseThreads")
    protected ExecutorService indexerWorkerPool = Executors.newFixedThreadPool(1);

    static final String MISSING_METADATA_MSG =
        "Metadata file is missing! Please, provide a metadata file! You may use \n" +
        "either 'metadata.yml', 'metadata.yaml' or 'metadata.json'!";

    public static final String GITLAB_MAIN_GROUP = "sharing";
    /**
     * list of standard branches, highest first
     */
    protected static final String[] SORTED_BRANCH_PRIORITY = { "main", "master" };

    protected static final String UNKNOWN_FILEPLACEHOLDER = "___UNRESOLVED_METADATAFILE___";

    protected JsonSchema metaDataSchema;

    @PostConstruct
    public void loadSchema() throws IOException {
        try (InputStream jsonSchemaResource = this.getClass().getResourceAsStream("metaData_elasticSearch.schema.json")) {
            JsonSchemaFactory factory = JsonSchemaFactory.getInstance(com.networknt.schema.SpecVersion.VersionFlag.V4);
            JsonSchema jsonSchema = factory.getSchema(jsonSchemaResource);
            metaDataSchema = jsonSchema;
        }
    }

    /**
     * infos about the complete commit, including project, branchname, commit hash
     * and list of errors.
     *
     * @author Micha
     *
     */
    static class AnalysedCommit {

        private final Project project;
        private final Commit commit;
        private final CollectionNode infoHierarchy;

        protected AnalysedCommit(Project project, Commit commit, CollectionNode infoHierarchy) {
            super();
            this.project = project;
            this.commit = commit;
            this.infoHierarchy = infoHierarchy;
        }

        public Project getProject() {
            return project;
        }

        public Commit getCommit() {
            return commit;
        }

        public CollectionNode getInfoHierarchy() {
            return infoHierarchy;
        }
    }

    /**
     * * just a derived class that does JSONignore exerciseId and ranking5 and
     * supportedActions,
     */
    protected static class ExtSearchResultDTO extends SearchResultDTO {

        /**
         * @return the exerciseId
         */
        @Override
        @JsonIgnore
        public String getExerciseId() {
            return super.getExerciseId();
        }

        /**
         * @return the ranking5
         */
        @Override
        @JsonIgnore
        public int getRanking5() {
            return super.getRanking5();
        }

        /**
         * @return the supportedActions
         */
        @Override
        @JsonIgnore
        public List<PluginActionInfo> getSupportedActions() {
            return super.getSupportedActions();
        }
    }

    public static class PublicVisibilityDeserializer extends JsonDeserializer<PublicVisibility> {

        @Override
        public PublicVisibility deserialize(JsonParser jsonParser, DeserializationContext deserializationContext) throws IOException {
            if (jsonParser.currentToken() == JsonToken.VALUE_TRUE) {
                PublicVisibility publicVisibility = new PublicVisibility();
                publicVisibility.setExcept(new String[] {});
                return publicVisibility;
            } else if (jsonParser.currentToken() == JsonToken.VALUE_FALSE) {
                return null;
            }
            return jsonParser.readValueAs(PublicVisibility.class);
        }
    }

    /**
     * just a derived class that does not JSONignore collectionContent. Public for
     * test purposes only.
     *
     * @author Michael
     *
     */
    public static class ExtUserProvidedMetadataDTO extends UserProvidedMetadataDTO {

        @JsonIgnore(false)
        private String[] collectionContent;

        @Override
        @SuppressWarnings("PMD.MethodReturnsInternalArray")
        public String[] getCollectionContent() {
            return collectionContent;
        }

        @Override
        @SuppressWarnings("PMD.ArrayIsStoredDirectly")
        public void setCollectionContent(String[] collectionContent) {
            this.collectionContent = collectionContent;
        }

        @JsonDeserialize(using = PublicVisibilityDeserializer.class)
        @Override
        @SuppressWarnings("PMD.UselessOverridingMethod")
        public PublicVisibility getPublicVisibility() {
            return super.getPublicVisibility();
        }

        @Override
        @SuppressWarnings("PMD.UselessOverridingMethod")
        public void setPublicVisibility(PublicVisibility publicVisibility) {
            super.setPublicVisibility(publicVisibility);
        }
    }

    /**
     * Represents a link to another Item (either relative to this path, to another
     * project,
     * or even to another repository
     *
     * @author Michael Breu
     *
     */
    public class ItemPath {

        public static final int MAX_ID_LENGTH = 512;

        // directory path either relative to some other Item path, or absolute to root
        // of object

        private final SlashPath path;

        /**
         * gitlab project id
         *
         */
        private final long projectId;

        private final Commit commit;

        public ItemPath(SlashPath path, Commit commit, Long projectId) {
            this.path = path;
            this.commit = commit;
            this.projectId = projectId;
        }

        public boolean isAbsolute() {
            return (path != null && path.startsWith("/"));
        }

        /**
         * """returns the full file path"""
         *
         * @return the full file path
         */
        public String getFullPath() {
            return path.toString();
        }

        @Override
        public int hashCode() {
            final int prime = 31;
            int result = 1;
            result = prime * result + Objects.hash(commit, path, projectId);
            return result;
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj) {
                return true;
            }
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            ItemPath other = (ItemPath) obj;
            return (Objects.equals(commit, other.commit) && Objects.equals(path, other.path) && projectId == other.projectId);
        }

        public SlashPath getPath() {
            return path;
        }

        public long getProjectId() {
            return projectId;
        }

        public Commit getCommit() {
            return commit;
        }

        public String getMetadataFile() {
            return this.path.getFileName().toString();
        }

        public String getDocId() {
            String prefix = "[" + this.projectId + "]";

            String pathString = this.path.getParent().toString();
            if (pathString == null || pathString.equals("/")) {
                return prefix;
            }
            int restLength = MAX_ID_LENGTH - prefix.length();
            if (restLength >= pathString.length()) {
                return prefix + pathString;
            } else {
                return prefix + pathString.substring(pathString.length() - restLength);
            }
        }

        public ItemPath createChildrenItemPath(String path, GitLabApi adminGitLabApi) throws GitLabApiException, ParseException {
            String[] parsedPath = parseChildPath(path);
            String gitUrl = parsedPath[0];
            long otherProjectId = parsedPath[1] == null ? -1L : Integer.parseInt(parsedPath[1]);
            String childPath = parsedPath[2];

            if (gitUrl == null) {
                if (otherProjectId < 0) {
                    SlashPath normalizedPath = normalizePath(childPath);
                    return new ItemPath(normalizedPath, this.commit, this.projectId);
                } else if (otherProjectId == projectId) {
                    if (childPath == null) {
                        return this;
                    }
                    SlashPath normalizedPath = SlashPath.get(childPath).normalize();
                    return new ItemPath(normalizedPath, this.commit, this.projectId);
                } else {
                    Project otherProject = adminGitLabApi.getProjectApi().getProject(otherProjectId);

                    String branchName = otherProject.getDefaultBranch();

                    Branch branch = adminGitLabApi.getRepositoryApi().getBranch(otherProject.getId(), branchName);
                    Commit lastCommit = branch.getCommit();
                    Commit otherCommit = new Commit();
                    otherCommit.setId(lastCommit.getId());
                    SlashPath normalizedPath = SlashPath.get(childPath == null ? "." : childPath).normalize();
                    return new ItemPath(normalizedPath, otherCommit, otherProject.getId());
                }
            } else {
                String message = String.format("Link %s to other repository not yet supported", gitUrl);
                logger.error(message);
                throw new ParseException(message, 0);
            }
        }

        public SlashPath normalizePath(String childPath) throws ParseException {
            if (childPath.contains("../")) {
                throw new ParseException(
                    "collectionContent path '" + childPath + "' contains a reference to a parent directory ('../'). This is not allowed.",
                    0
                );
            }
            if (this.getPath().isDirectory()) {
                SlashPath resolvedWrtDirectory = this.getPath().resolve(childPath).normalize();
                if (childPath.endsWith("/")) {
                    resolvedWrtDirectory = resolvedWrtDirectory.resolve(UNKNOWN_FILEPLACEHOLDER);
                }
                return resolvedWrtDirectory;
            }

            final SlashPath parent = this.getPath().getParent();
            if ("/".equals(parent.toString())) {
                // we avoid prefixed roots
                return SlashPath.get(childPath);
            }
            return parent.resolve(childPath).normalize();
        }

        @Override
        public String toString() {
            return "ItemPath[" + path + "]";
        }
    }

    protected static String[] parseChildPath(String childPath) throws ParseException {
        Matcher match = collectionContentPattern.matcher(childPath);
        if (!match.find()) {
            throw new ParseException("Cannot parse " + childPath, 0);
        }
        String childGitUrl = match.group(1);
        String childProjectId = null;
        if (!StringUtils.isEmpty(match.group(3))) {
            childProjectId = match.group(3);
        }
        String childPathString = match.group(4);
        if (childProjectId != null && childPathString != null && childPathString.startsWith("/")) {
            // we transfer it into a relative path
            childPathString = childPathString.substring(1);
        }
        return new String[] { childGitUrl, childProjectId, childPathString };
    }

    public class HealthCheck {

        protected ObjectMapper yamlMapper = new ObjectMapper(new YAMLFactory());

        protected ObjectMapper jsonMapper = new ObjectMapper(new JsonFactory());

        protected HealthCheck() {
            super();
            yamlMapper.findAndRegisterModules();
            jsonMapper.findAndRegisterModules();
        }

        public Optional<CollectionNode> validateProject(Project project, Commit commit) throws GitLabApiException {
            return validateFile(new ItemPath(SlashPath.get("."), commit, project.getId()), null, project);
        }

        private Optional<CollectionNode> validateFile(ItemPath path, CollectionNode parent, Project project) throws GitLabApiException {
            if (parent != null && path.projectId != parent.getGitLabProject().getId()) {
                MetadataInfo metadataFile = new MetadataInfo(null, path.getFullPath());
                var metadataInfo = new MetadataInfo(null, metadataFile.getFilename());
                metadataInfo.getWarnings().add("children in other git projects are not validated or updated");
                return Optional.of(new CollectionNode(path, metadataInfo, null, parent, project));
            }
            String metadataFilePath;
            if (path.path.isDirectory()) {
                List<String> metadataFiles = getRootRepositoryMetadataFiles(path);
                MetadataInfo metadataInfo = checkForSingleMetadataFiles(metadataFiles);
                if (!metadataInfo.getErrors().isEmpty()) {
                    return Optional.of(new CollectionNode(path, metadataInfo, null, parent, project));
                }
                metadataFilePath = metadataFiles.get(0);
            } else {
                metadataFilePath = path.path.toString();
            }

            MetadataInfo metadataFile = new MetadataInfo(null, metadataFilePath);

            ExtUserProvidedMetadataDTO metaData;
            try {
                metaData = parseMetadata(path.getProjectId(), path.commit, metadataFile);
            } catch (ParseException e) {
                var metadataInfo = new MetadataInfo(null, metadataFile.getFilename());
                metadataInfo.getErrors().add(e.getMessage());
                return Optional.of(new CollectionNode(path, metadataInfo, null, parent, project));
            }

            var metadataInfo = new MetadataInfo(null, metadataFilePath);
            metadataInfo.setMetaData(metaData);
            CollectionNode node = new CollectionNode(
                new ItemPath(SlashPath.get(metadataFile.getFilename()), path.commit, path.getProjectId()),
                metadataInfo,
                metadataFile.getRepositoryFile(),
                parent,
                project
            );
            return validateMetadataFile(node);
        }

        @SuppressWarnings("PMD.PreserveStackTrace")
        private ExtUserProvidedMetadataDTO parseMetadata(long projectId, Commit commit, MetadataInfo metadataFile) throws ParseException {
            try {
                String filename = metadataFile.getFilename();
                // remove leading slashes
                while (filename.startsWith("/")) {
                    filename = filename.substring(1);
                }
                final RepositoryFile mdFile = adminGitLabApi.getRepositoryFileApi().getFile(projectId, filename, commit.getId());
                String content = new String(Base64.getDecoder().decode(mdFile.getContent()), StandardCharsets.UTF_8);
                var mapper = mdFile.getFileName().endsWith(".json") ? jsonMapper : yamlMapper;

                metadataFile.setRepositoryFile(mdFile);
                return mapper.readValue(content, ExtUserProvidedMetadataDTO.class);
            } catch (GitLabApiException | JsonProcessingException e) {
                logger.warn("Parseexception?", e);
                throw new ParseException(e.getMessage(), 0);
            }
        }

        private Optional<CollectionNode> validateMetadataFile(CollectionNode node) {
            return validateMetadataFile(node, new ArrayList<>());
        }

        @SuppressWarnings("PMD.AvoidThrowingNullPointerException")
        private Optional<CollectionNode> validateMetadataFile(CollectionNode node, List<ItemPath> parents) {
            ItemPath thisPath = node.getItem();
            if (parents.contains(thisPath)) {
                node.getMetaDataInfo().getErrors().add("This item has its own parent?");
                return Optional.empty();
            }
            if (node.getMetaDataInfo().getMetaData().isEmpty()) {
                throw new NullPointerException("metadata should be successfully parsed here");
            }

            final ValidationResultDTO validationResult = vocabularyService.validateMetaData(
                node.getMetaDataInfo().getMetaData().get(),
                node.isRoot()
            );
            furtherValidations(validationResult, node.getMetaDataInfo().getMetaData().get());
            node.getMetaDataInfo().getErrors().addAll(validationResult.getErrors());
            node.getMetaDataInfo().getWarnings().addAll(validationResult.getWarnings());

            final String[] collectionContent = node.getMetaDataInfo().getMetaData().get().getCollectionContent();

            if (collectionContent != null && collectionContent.length != 0) {
                List<ItemPath> thisParents = new ArrayList<>(parents);
                thisParents.add(thisPath);
                Arrays
                    .stream(collectionContent)
                    .map(childName -> {
                        try {
                            return node.getItem().createChildrenItemPath(childName, adminGitLabApi);
                        } catch (GitLabApiException | ParseException e1) {
                            logger.error("Cannot parse child {}", childName, e1);
                            node.getMetaDataInfo().getErrors().add("Cannot parse child " + childName + ": " + e1.getMessage());
                            return null;
                        }
                    })
                    .forEach(childPath -> {
                        try {
                            final Optional<CollectionNode> subValidationResult = validateFile(childPath, node, node.getGitLabProject());
                            subValidationResult.ifPresent(r -> node.getChildren().add(r));
                        } catch (GitLabApiException e1) {
                            logger.error("Cannot parse child {}", childPath, e1);
                            node.getMetaDataInfo().getErrors().add("Cannot parse child " + childPath.toString() + ": " + e1.getMessage());
                            return;
                        }
                    });
            }

            return Optional.of(node);
        }

        /**
         * adds further validations
         * @param validationResult
         * @param extUserProvidedMetadataDTO
         */
        private void furtherValidations(ValidationResultDTO validationResult, ExtUserProvidedMetadataDTO extUserProvidedMetadataDTO) {
            // check version
            if (extUserProvidedMetadataDTO.getMetadataVersion() == null || !extUserProvidedMetadataDTO.getMetadataVersion().equals("0.4")) {
                validationResult.addError("Currently only metadata version 0.4 is supported");
            }
            // TODO: further checks
        }

        public List<String> getRootRepositoryMetadataFiles(ItemPath item) throws GitLabApiException {
            if (item.path.isDirectory()) {
                return getRepositoryMetadataFiles(item.getProjectId(), item.getCommit(), item.getPath().toString());
            } else {
                return Collections.singletonList(item.getPath().toString());
            }
        }

        public List<String> getRepositoryMetadataFiles(long projectId, Commit commit, String path) throws GitLabApiException {
            final Stream<TreeItem> files = adminGitLabApi
                .getRepositoryApi()
                .getTreeStream(projectId, path, commit.getId(), false/* only this folder */);
            return files.filter(f -> METADATA_FILENAMES.contains(f.getName())).map(f -> f.getPath()).collect(Collectors.toList());
        }

        public MetadataInfo checkForSingleMetadataFiles(List<String> metadataFiles) {
            MetadataInfo metadataInfo = new MetadataInfo(null, null);
            int numMetadataFiles = metadataFiles.size();
            if (numMetadataFiles == 0) {
                String errorMsg = MISSING_METADATA_MSG;
                metadataInfo.getErrors().add(errorMsg);
                metadataInfo.setFilename("Metadata file is missing!");
            } else if (numMetadataFiles > 1) {
                String errorMsg =
                    "Multiple metadata files were found! Please, choose one of the currently provided files: " +
                    String.join(", ", metadataFiles);
                metadataInfo.getErrors().add(errorMsg);
                metadataInfo.setFilename("Multiple metadata files found!");
            }
            return metadataInfo;
        }
    }

    public ProjectPermissions getPermissionsfromProject(Project p) {
        ProjectPermissions result = new ProjectPermissions();
        adminGitLabApi.enableRequestResponseLogging(java.util.logging.Level.INFO);
        try {
            for (Member member : adminGitLabApi.getProjectApi().getMembers(p.getId())) {
                if (member.getEmail() == null) {
                    // member does not show email address?
                    User user = adminGitLabApi.getUserApi().getUser(member.getId());
                    result.emailAddressesWithAccess.add(user.getEmail());
                } else {
                    result.emailAddressesWithAccess.add(member.getEmail());
                }
            }
        } catch (GitLabApiException e) {
            logger.error("Cannot access members for {} ({})", p.getNameWithNamespace(), p.getId(), e);
        }

        final List<ProjectSharedGroup> sharedWithGroups = p.getSharedWithGroups();
        for (ProjectSharedGroup group : sharedWithGroups) {
            result.groupsWithAccess.add(group.getGroupName());
        }
        return result;
    }

    public class Indexing {

        private final String index;

        protected Indexing(String index) {
            super();
            this.index = index;
        }

        public void indexAllMetadata(ProjectPermissions projectPermissions, CollectionNode metadataTree) {
            indexMetadataNode(projectPermissions, metadataTree);
        }

        protected GitProjectDTO mapProject(Project p, ProjectPermissions perm) {
            final String[] split = p.getPathWithNamespace().split("/");
            String mainGroup = split[0];
            String subGroup = split.length > 2 ? split[1] : null;

            GitProjectDTO gp = new GitProjectDTO();
            gp.setLast_activity_at(p.getLastActivityAt().toInstant());
            gp.setMain_group(mainGroup);
            gp.setNamespace(p.getPathWithNamespace());
            gp.setProject_id(p.getId().intValue());
            gp.setProject_name(p.getName());
            gp.setSub_group(subGroup);
            gp.setUrl(p.getWebUrl());

            // extra properties

            gp.setArchived(p.getArchived());
            gp.setVisibility(p.getVisibility());
            gp.setUsers(perm.getEmailAddressesWithAccess());
            gp.setGroups(perm.getGroupsWithAccess());
            gp.setOpen_issues_count(p.getOpenIssuesCount());
            gp.setForks_count(p.getForksCount());
            gp.setLast_activity_at(p.getLastActivityAt().toInstant());
            gp.setDescription(p.getDescription());

            return gp;
        }

        protected MetadataFile mapFile(RepositoryFile rFile, String parentId, String[] childIds) {
            MetadataFile file = new MetadataFile();
            file.setCommit_id(rFile.getCommitId());
            file.setFilename(rFile.getFileName());
            file.setPath(rFile.getFilePath());
            file.setIndexing_date(Instant.now());
            file.setLast_activity_at(Instant.now());
            file.setParentId(parentId);
            file.setChildren(childIds);

            return file;
        }

        protected ExtSearchResultDTO constructESDocument(ProjectPermissions projectPermissions, CollectionNode node) {
            if (node.getMetaDataInfo().getMetaData().isEmpty()) {
                return null;
            }
            Optional<String> parentId = node.getParent() == null ? Optional.empty() : Optional.of(node.getParent().getItem().getDocId());
            ExtSearchResultDTO doc = new ExtSearchResultDTO();
            doc.setProject(mapProject(node.getGitLabProject(), projectPermissions));
            final String[] childIds = node
                .getChildren()
                .stream()
                .map(childNode -> childNode.getItem().getDocId())
                .collect(Collectors.toList())
                .toArray(new String[] {});

            final MetadataFile mapFile = mapFile(node.getRepositoryFile(), parentId.orElse(null), childIds);

            doc.setFile(mapFile);
            doc.setExerciseId(node.getItem().getDocId());
            doc.setMetadata(node.getMetaDataInfo().getMetaData().get());

            final String interactivityType = doc.getMetadata().getInteractivityType();
            doc.getMetadata().setInteractivityType(interactivityType);
            //            if (!isEnumInteractivity(interactivityType)) {
            //                // logger.warn("TODO: Reset clearing of InteractivityType! Was just for
            //                // temporary backward compatibility");
            //                doc.getMetadata().setInteractivityType(InteractivityTypeXXX.UNSPECIFIED.getExternalName());
            //            }

            esRepository.updateSearchStatistics(doc);
            return doc;
        }

        @SuppressWarnings({ "deprecation", "PMD.UnusedPrivateMethod", "unused" })
        private boolean isEnumInteractivity(String s) {
            if (s == null) {
                return true;
            }
            for (InteractivityTypeXXX ia : InteractivityTypeXXX.values()) {
                if (ia.getExternalName().equals(s)) {
                    return true;
                }
            }
            return false;
        }

        /**
         * indexes all nodes in the collection tree, and removes all outdated indices.
         *
         * @param projectPermissions permissions for project
         * @param metadataTree       the collection tree
         * @return the set of created id's
         */
        public List<String> indexMetadataNode(ProjectPermissions projectPermissions, CollectionNode metadataTree) {
            List<String> existingDocIds = new ArrayList<>();
            try {
                existingDocIds = esRepository.getIdsByProjectId(metadataTree.getGitLabProject().getId());
            } catch (IOException e) {
                logger.error("Cannot find exsting ids for project {}", metadataTree.getGitLabProject().getId());
            }
            List<String> newDocIds = indexMetadataNodeRecursive(projectPermissions, metadataTree);

            existingDocIds
                .stream()
                .filter(id -> !newDocIds.contains(id))
                .forEach(id -> {
                    try {
                        esRepository.deleteById(id);
                    } catch (IOException e) {
                        logger.error("Cannot delete document for id {}", id, e);
                    }
                });
            return newDocIds;
        }

        /**
         * indexes all nodes in the collection tree
         *
         * @param projectPermissions permissions for project
         * @param metadataTree       the collection tree
         * @return the set of created id's
         */
        private List<String> indexMetadataNodeRecursive(ProjectPermissions projectPermissions, CollectionNode metadataTree) {
            var docId = metadataTree.getItem().getDocId();
            List<String> result = new ArrayList<>();
            logger.debug(
                "Indexing metadata for {} of project {} in {}",
                docId,
                metadataTree.getGitLabProject().getNameWithNamespace(),
                metadataTree.getItem().getFullPath()
            );

            try {
                final ExtSearchResultDTO constructESDocument = constructESDocument(projectPermissions, metadataTree);
                if (constructESDocument != null) {
                    result.add(esRepository.index(constructESDocument, index));
                }
            } catch (IOException e) {
                logger.error("Cannot index document for {}", metadataTree.getItem().getDocId(), e);
            }

            for (CollectionNode child : metadataTree.getChildren()) {
                result.addAll(indexMetadataNodeRecursive(projectPermissions, child));
            }
            return result;
        }

        public long deleteProjectFromIndex(long projectId) {
            try {
                return esRepository.deleteByProjectId(projectId);
            } catch (IOException e) {
                logger.error("Cannot delete all from project {}", projectId, e);
                return 0;
            }
        }
    }

    protected static final List<String> METADATA_FILENAMES = Collections.unmodifiableList(
        new ArrayList<String>() {
            private static final long serialVersionUID = 1L;

            {
                add("metadata.yaml");
                add("metadata.yml");
                add("metadata.json");
            }
        }
    );

    @PostConstruct
    public void setAdminApi() {
        adminGitLabApi = gitLabRepository.getAdminGitLabApi();
        if (adminGitLabApi == null) {
            logger.error("adminGitLabApi is null??");
        } else {
            logger.debug("gitlab url: {}", adminGitLabApi.getGitLabServerUrl());
        }
    }

    /**
     * Waits for the previous task to be finished. Mainly for test purposes.
     */
    @SuppressWarnings("PMD.DoNotUseThreads")
    public void waitForFinish() {
        try {
            Callable<Object> callable = () -> {
                logger.info("previous task finished");
                return null;
            };
            final Future<Object> submitFuture = indexerWorkerPool.submit(callable);
            submitFuture.get(10, TimeUnit.MINUTES);
        } catch (TimeoutException | ExecutionException | InterruptedException e) {
            logger.warn("Wait got interrupted?", e);
        }
    }

    @SuppressWarnings({ "PMD.CognitiveComplexity", "PMD.CyclomaticComplexity", "PMD.DoNotUseThreads" })
    public void handleGitLabEvent(SystemHookEvent gitLabEvent) {
        @SuppressWarnings("PMD.AvoidCatchingGenericException")
        Runnable r = () -> {
            try {
                if (gitLabEvent instanceof PushSystemHookEvent) {
                    handlePushEvent((PushSystemHookEvent) gitLabEvent);
                } else if (gitLabEvent instanceof ProjectSystemHookEvent) {
                    ProjectSystemHookEvent projectEvent = (ProjectSystemHookEvent) gitLabEvent;
                    if ("project_rename".equals(projectEvent.getEventName())) {
                        handleProjectRenameEvent(projectEvent);
                    }
                    if ("project_transfer".equals(projectEvent.getEventName())) {
                        handleProjectRenameEvent(projectEvent);
                    }
                    if ("project_destroy".equals(projectEvent.getEventName())) {
                        handleProjectDeleteEvent(projectEvent);
                    } else {
                        logger.warn("Project Event {} not handled", gitLabEvent.getEventName());
                    }
                } else if (gitLabEvent instanceof GroupSystemHookEvent) {
                    GroupSystemHookEvent groupEvent = (GroupSystemHookEvent) gitLabEvent;
                    if ("group_rename".equals(groupEvent.getEventName())) {
                        handleGroupRenameEvent(groupEvent);
                    } else {
                        logger.warn("Group Event {} not handled", gitLabEvent.getEventName());
                    }
                } else if (gitLabEvent instanceof TeamMemberSystemHookEvent) {
                    TeamMemberSystemHookEvent userEvent = (TeamMemberSystemHookEvent) gitLabEvent;
                    if ("user_add_to_team".equals(userEvent.getEventName()) || "user_remove_from_team".equals(userEvent.getEventName())) {
                        handleTeamChangeEvent(userEvent);
                    } else {
                        logger.warn("Team Event {} not handled", gitLabEvent.getEventName());
                    }
                } else {
                    if ("repository_update".equals(gitLabEvent.getEventName())) {
                        logger.debug("Ignoring event {}", gitLabEvent.getEventName());
                    } else if ("user_add_to_group".equals(gitLabEvent.getEventName())) {
                        logger.debug("Ignoring event {}", gitLabEvent.getEventName());
                    } else {
                        logger.error("Cannot handle event {}", gitLabEvent.getEventName());
                        throw new MethodNotFoundException(gitLabEvent.getEventName(), null);
                    }
                }
            } catch (GitLabApiException | IOException e) {
                logger.error("Cannot process gitLab event {} ({})", gitLabEvent.getEventName(), gitLabEvent.getRequestUrl(), e);
            }
        };
        indexerWorkerPool.execute(r);
    }

    protected AnalysedCommit handleProjectRenameEvent(ProjectSystemHookEvent projectRenameEvent) throws GitLabApiException, IOException {
        logger.info("rename project {}({})", projectRenameEvent.getName(), projectRenameEvent.getProjectId());
        final Project project = adminGitLabApi.getProjectApi().getProject(projectRenameEvent.getProjectId());
        handleProjectDeleteEvent(projectRenameEvent);
        indexProject(project);
        logger.info("rename project {}({}) finished", projectRenameEvent.getName(), projectRenameEvent.getProjectId());
        // no mail sending
        return null;
    }

    protected AnalysedCommit handleProjectDeleteEvent(ProjectSystemHookEvent projectEvent) throws IOException {
        logger.info("handle project destroy {}", projectEvent.getProjectId());
        final Long deleteCount = esRepository.deleteByProjectId(projectEvent.getProjectId());
        logger.info("handled project destroy {}: {} entries deleted", projectEvent.getProjectId(), deleteCount);
        return null;
    }

    protected AnalysedCommit handleGroupRenameEvent(GroupSystemHookEvent groupEvent) throws GitLabApiException, IOException {
        logger.info("handle group rename for {} -> {}", groupEvent.getOldFullPath(), groupEvent.getFullPath());
        final List<String> oldIds = esRepository.getIdsByNamespacePrefix(groupEvent.getOldFullPath());
        oldIds.forEach(id -> {
            try {
                esRepository.deleteById(id);
                logger.info("Deleted old id {}", id);
            } catch (IOException e) {
                logger.info("Deleted old id {} failed", id, e);
            }
        });

        gitLabRepository
            .streamAllProjects()
            .filter(project -> project.getPathWithNamespace().startsWith(groupEvent.getFullPath()))
            .forEach(project -> indexProject(project));

        return null;
    }

    private AnalysedCommit handleTeamChangeEvent(TeamMemberSystemHookEvent teamEvent) throws GitLabApiException {
        logger.error("handle team change for project {}({})", teamEvent.getProjectName(), teamEvent.getProjectId());
        final Project project = adminGitLabApi.getProjectApi().getProject(teamEvent.getProjectId());
        indexProject(project);
        logger.error("handle team change for project {}({}) finished", teamEvent.getProjectName(), teamEvent.getProjectId());
        return null;
    }

    protected AnalysedCommit handlePushEvent(PushSystemHookEvent pushEvent) throws GitLabApiException {
        boolean metaDataMandatory = isInMainGroup(pushEvent.getProject().getPathWithNamespace());
        logger.info(
            "Indexing {} ({}) is mandatory: {}",
            pushEvent.getProject().getPathWithNamespace(),
            pushEvent.getProjectId(),
            metaDataMandatory
        );
        String branchName = pushEvent.getBranch();
        if (Arrays.binarySearch(SORTED_BRANCH_PRIORITY, branchName) < 0) {
            return null;
        }
        Long projectId = pushEvent.getProjectId();
        Project project = adminGitLabApi.getProjectApi().getProject(projectId);
        final Commit commit = adminGitLabApi.getCommitsApi().getCommit(projectId, pushEvent.getAfter());
        final AnalysedCommit handledCommit = handleCommit(metaDataMandatory, project, commit);
        sendCommitMailWhenErrorsOrWarnings(pushEvent.getUserEmail(), handledCommit);
        return handledCommit;
    }

    /**
     * handles the comit
     *
     * @param metaDataMandatory
     * @param branchName
     * @param project
     * @param commit
     * @return
     */
    private AnalysedCommit handleCommit(boolean metaDataMandatory, Project project, final Commit commit) {
        AnalysedCommit analyzedCommit = checkAndIndexProject(project, commit, metaDataMandatory);
        if (analyzedCommit == null) {
            logger.info("Stopped handling push event. Error occurred while handling push event.");
        }
        return analyzedCommit;
    }

    protected boolean isInMainGroup(String pathWithNamespace) {
        return pathWithNamespace.startsWith(GITLAB_MAIN_GROUP + "/");
    }

    public AnalysedCommit checkAndIndexProject(Project project, Commit commitHash, boolean metadataMandatory) {
        return checkAndIndexProject(project, commitHash, metadataMandatory, SearchRepositoryConstants.INDEX_METADATA);
    }

    @SuppressWarnings({ "PMD.AvoidCatchingGenericException", "PMD.CyclomaticComplexity" })
    public AnalysedCommit checkAndIndexProject(Project project, Commit eventCommit, boolean metadataMandatory, String indexName) {
        HealthCheck healthCheck = new HealthCheck();
        Optional<CollectionNode> tree;
        try {
            tree = healthCheck.validateProject(project, eventCommit);
        } catch (Exception e) {
            logger.error("Exception during healthcheck", e);
            tree = Optional.empty();
        }
        if (tree.isEmpty()) {
            logger.info("project {} is not index. No metadata found", project.getId());
            return null;
        }
        boolean isMetadataMissing = false;

        // Missing on top level?
        for (String error : tree.get().getMetaDataInfo().getErrors()) {
            if (MISSING_METADATA_MSG.equals(error)) {
                isMetadataMissing = true;
                break;
            }
        }
        if ((isMetadataMissing || tree.get().getMetaDataInfo().getMetaData().isEmpty()) && !metadataMandatory) {
            logger.info("project {} is not index. No metadata found", project.getId());
            return null;
        }
        if (tree.get().hasErrors()) {
            String errors = tree
                .get()
                .getErrors()
                .entrySet()
                .stream()
                .map(e -> String.format("%s: %s", e.getKey(), e.getValue().stream().collect(Collectors.joining(", "))))
                .collect(Collectors.joining("\\n"));
            logger.info("project {} has errors and is not indexed: {}", project.getId(), errors);
        } else {
            if (logger.isDebugEnabled()) {
                logger.debug("parsed tree:");
                tree.get().printTreeRecursive();
            }

            Indexing indexing = new Indexing(indexName);
            indexing.indexMetadataNode(getPermissionsfromProject(project), tree.get());
        }

        return new AnalysedCommit(project, eventCommit, tree.get());
    }

    @SuppressWarnings({ "PMD.CloseResource", "PMD.DoNotUseThreads" })
    public LogHolder indexAllProjectsAsync() throws IOException {
        LoggerContext context = (LoggerContext) LoggerFactory.getILoggerFactory();
        ch.qos.logback.classic.Logger rootLogger = (ch.qos.logback.classic.Logger) LoggerFactory.getLogger(logger.getName());

        final PipedOutputStream pos = new PipedOutputStream(); // we have to close it in other thread
        final PipedInputStream pis = new PipedInputStream(pos, 50000);

        final LogHolder log = new LogHolder(pis);

        final OutputStreamAppender<ILoggingEvent> outAppender = new OutputStreamAppender<>();
        outAppender.setContext(context);
        final PatternLayoutEncoder encoder = new PatternLayoutEncoder();
        encoder.setPattern("%d{HH:mm:ss.SSS} %-4relative %-5level %logger{35} - %msg%n");
        encoder.setContext(context);
        encoder.start();
        outAppender.setEncoder(encoder);
        outAppender.setName("WebSocketLogger");
        outAppender.setOutputStream(pos);

        rootLogger.addAppender(outAppender);
        outAppender.start();

        @SuppressWarnings("PMD.AvoidCatchingGenericException")
        Runnable r = () -> {
            try {
                logger.info("Indexing all projects");
                indexAllProjects();
                logger.info("Indexing finished!");
                log.setClosed(true);
            } catch (Exception t) {
                logger.error("reindexing was closed unexpectedly", t);
                log.setClosed(true);
                log.setError(true);
                outAppender.stop();
                try {
                    pos.close();
                } catch (IOException e) {
                    logger.error("pos cannot be closed", e);
                }
            }
        };

        indexerWorkerPool.execute(r);

        return log;
    }

    public void indexAllProjects() throws GitLabApiException, IOException {
        String index = esRepository.createNewMetaDataIndex();

        gitLabRepository.streamAllProjects().forEach(project -> indexProject(index, project));
        esRepository.createAlias(index);
    }

    private void indexProject(Project project) {
        indexProject(SearchRepositoryConstants.INDEX_METADATA, project);
    }

    private void indexProject(String index, Project project) {
        boolean metaDataMandatory = isInMainGroup(project.getPathWithNamespace());
        try {
            Optional<Branch> branch = getBranchToIndex(project);
            if (branch.isPresent()) {
                logger.info(
                    "Indexing project {} [{}/{}({})]",
                    project.getNameWithNamespace(),
                    project.getId(),
                    branch.get().getName(),
                    metaDataMandatory
                );
                Commit commit = new Commit();
                commit.setId(branch.get().getCommit().getId());
                AnalysedCommit analyzedCommit = checkAndIndexProject(project, commit, metaDataMandatory, index);
                if (analyzedCommit == null) {
                    logger.info(
                        "Stopped handling push event for {}/{}. Error occurred while handling push event.",
                        project.getId(),
                        commit.getId()
                    );
                }
            }
        } catch (GitLabApiException e) {
            logger.warn("Cannot find branch or commit in project {}", project.getNameWithNamespace(), e);
        }
    }

    /**
     *
     * Returns the name of the branch which should be indexed,
     * following the priority given by BRANCH_PRIORITY.
     * Raises NoBranchToIndexError if no branch with a name
     * appearing in BRANCH_PRIORITY exists.
     *
     * @param project The project whose branches should be checked
     * @throws GitLabApiException
     * @returns The name of the branch to index, or Optional.empty() if not found.
     *
     *          """
     *
     */
    protected Optional<Branch> getBranchToIndex(Project project) throws GitLabApiException {
        final List<Branch> projectBranches = adminGitLabApi.getRepositoryApi().getBranches(project.getId());
        return projectBranches.stream().filter(branch -> Arrays.binarySearch(SORTED_BRANCH_PRIORITY, branch.getName()) >= 0).findAny();
    }

    /**
     * Sends an email to the user who triggered the validation warning about errors.
     * Only call this method if there were errors, if the validation did not find
     * any issues no email should be sent
     *
     * @param git_event       the event which triggered the validation
     * @param analysed_commit information about the commit which was analysed
     *
     * @param commit
     */
    @SuppressWarnings("PMD.ConfusingTernary")
    public void sendCommitMailWhenErrorsOrWarnings(String defaultEmail, AnalysedCommit commit) {
        if (commit != null && (commit.getInfoHierarchy().hasErrors() || commit.getInfoHierarchy().hasWarnings())) {
            String email = defaultEmail;
            if (StringUtils.isEmpty(email)) {
                email = commit.getCommit().getCommitterEmail();
                if (StringUtils.isEmpty(email)) {
                    email = commit.getCommit().getAuthorEmail();
                }
            }
            if (StringUtils.isEmpty(email)) {
                logger.warn("Cannot send mail to '{}' for {}/{}", email, commit.getProject().getId(), commit.getCommit().getId());
            } else {
                final String mainMessage = String.format(
                    "Sharing Platform: Project %s -> %s",
                    commit.getProject().getPathWithNamespace(),
                    (
                        commit.infoHierarchy.getErrors().isEmpty()
                            ? "repository health check returned warnings!"
                            : "repository health check failed!"
                    )
                );

                Locale locale = Locale.forLanguageTag("en");
                Context context = new Context(locale);
                context.setVariable("commit", commit);
                context.setVariable("mainMessage", mainMessage);
                context.setVariable(MailService.BASE_URL, jHipsterProperties.getMail().getBaseUrl());

                logger.info("Sending commit mail to {}", email);
                String content = templateEngine.process("mail/gitlab_events/indexingResult", context);

                @SuppressWarnings("unused")
                Boolean success = mailService.sendEmail(email, mainMessage, content, false, true);
            }
        }
    }
}
